#!/usr/bin/perl

#
# "Disassembler" (C) 30-Jul-2020y by *HW*
#
our $code_page = 8;

sub xcode {
    my ($str) = @_;
    my ($result, $ch) = ("");

    for $ch (split //, $str) {
        $result .= $ch, next if ord($ch) < 0x60;
        $result .= '?', next if ord($ch) > 0x7F;
        if ($code_page == 7) { # KOI-7
            $result .= $ch;
        }
        elsif ($code_page == 8) { # KOI-8
            $result .= chr(ord($ch) + 0x80);
        }
        elsif ($code_page == 866) { # OEM CP 866
            $result .= substr(
                "\x9E\x80\x81\x96\x84\x85\x94\x83".
                "\x95\x88\x89\x8A\x8B\x8C\x8D\x8E".
                "\x8F\x9F\x90\x91\x92\x93\x86\x82".
                "\x9C\x9B\x87\x98\x9D\x99\x97\x9A",
                ord($ch) - 0x60, 1
            );
        }
        elsif ($code_page == 1251) { # ANSI CP Windows 1251
            $result .= substr(
                "\xDE\xC0\xC1\xD6\xC4\xC5\xD4\xC3".
                "\xD5\xC8\xC9\xCA\xCB\xCC\xCD\xCE".
                "\xCF\xDF\xD0\xD1\xD2\xD3\xC6\xC2".
                "\xDC\xDB\xC7\xD8\xDD\xD9\xD7\xDA",
                ord($ch) - 0x60, 1
            );
        }
        else {
            $result .= '?';
        }
    }
    return $result;
}

sub ccode {
    my ($str) = @_;
    my ($result, $ch, $qo, $qc, $mode) = (
        "", '_', "'", "'", 0
    );

    for $ch (split //, $str) {
        if (ord($ch) < 0x20 or $ch eq $qo or $ch eq $qc) {
            $result .= "$qc, " if $mode == 2;
            $result .= ", " if $mode == 1;
            $result .= ord($ch);
            $mode = 1;
        }
        else {
            $result .= "$qo" if $mode == 0;
            $result .= ", $qo" if $mode == 1;
            $result .= $ch;
            $mode = 2;
        }
    }
    $result .= "$qc" if $mode == 2;
    return $result;
}

sub dcode {
    my ($str) = @_;
    $str =~ tr{\x00-\x1F}{.};
    return $str;
}

#------------------------------------------------------------------------------
# Instruction tables are from
# https://github.com/lllllzzzzz/Intel-8080-Disassembler/blob/master/disassembler.c
# They contained a lot of bugs, those I found I fixed
#

@INSTR_SZ = (
#   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
    1, 3, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1,
    1, 3, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1,
    1, 3, 3, 1, 1, 1, 2, 1, 1, 1, 3, 1, 1, 1, 2, 1,
    1, 3, 3, 1, 1, 1, 2, 1, 1, 1, 3, 1, 1, 1, 2, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 3, 3, 3, 1, 2, 1, 1, 1, 3, 1, 3, 3, 2, 1,
    1, 1, 3, 2, 3, 1, 2, 1, 1, 1, 3, 2, 3, 1, 2, 1,
    1, 1, 3, 1, 3, 1, 2, 1, 1, 1, 3, 1, 3, 3, 2, 1,
    1, 1, 3, 1, 3, 1, 2, 1, 1, 1, 3, 1, 3, 3, 2, 1
);
@INSTR = (
#     0          1          2          3          4          5           6          7          8          9          A          b          c          d          e          f
    "nop",     "lxi b,",  "stax b",  "inx b",   "inr b",   "dcr b",    "mvi b,",  "rlc",     "illegal", "dad b",   "ldax b",  "dcx b",   "inr c",   "dcr c",   "mvi c,",  "rrc",
    "illegal", "lxi d,",  "stax d",  "inx d",   "inr d",   "dcr d",    "mvi d,",  "ral",     "illegal", "dad d",   "ldax d",  "dcx d",   "inr e",   "dcr e",   "mvi e,",  "rar",
    "illegal", "lxi h,",  "shld",    "inx h",   "inr h",   "dcr h",    "mvi h,",  "daa",     "illegal", "dad h",   "lhld",    "dcx h",   "inr l",   "dcr l",   "mvi l,",  "cma",
    "illegal", "lxi sp,", "sta",     "inx sp",  "inr m",   "dcr m",    "mvi m,",  "stc",     "illegal", "dad sp",  "lda",     "dcx sp",  "inr a",   "dcr a",   "mvi a,",  "cmc",
    "mov b,b", "mov b,c", "mov b,d", "mov b,e", "mov b,h", "mov b,l",  "mov b,m", "mov b,a", "mov c,b", "mov c,c", "mov c,d", "mov c,e", "mov c,h", "mov c,l", "mov c,m", "mov c,a",
    "mov d,b", "mov d,c", "mov d,d", "mov d,e", "mov d,h", "mov d,l",  "mov d,m", "mov d,a", "mov e,b", "mov e,c", "mov e,d", "mov e,e", "mov e,h", "mov e,l", "mov e,m", "mov e,a",
    "mov h,b", "mov h,c", "mov h,d", "mov h,e", "mov h,h", "mov h,l",  "mov h,m", "mov h,a", "mov l,b", "mov l,c", "mov l,d", "mov l,e", "mov l,h", "mov l,l", "mov l,m", "mov l,a",
    "mov m,b", "mov m,c", "mov m,d", "mov m,e", "mov m,h", "mov m,l",  "hlt",     "mov m,a", "mov a,b", "mov a,c", "mov a,d", "mov a,e", "mov a,h", "mov a,l", "mov a,m", "mov a,a",
    "add b",   "add c",   "add d",   "add e",   "add h",   "add l",    "add m",   "add a",   "adc b",   "adc c",   "adc d",   "adc e",   "adc h",   "adc l",   "adc m",   "adc a",
    "sub b",   "sub c",   "sub d",   "sub e",   "sub h",   "sub l",    "sub m",   "sbb a",   "sbb b",   "sbb c",   "sbb d",   "sbb e",   "sbb h",   "sbb l",   "sbb m",   "sbb a",
    "ana b",   "ana c",   "ana d",   "ana e",   "ana h",   "ana l",    "ana m",   "ana a",   "xra b",   "xra c",   "xra d",   "xra e",   "xra h",   "xra l",   "xra m",   "xra a",
    "ora b",   "ora c",   "ora d",   "ora e",   "ora h",   "ora l",    "ora m",   "ora a",   "cmp b",   "cmp c",   "cmp d",   "cmp e",   "cmp h",   "cmp l",   "cmp m",   "cmp a",
    "rnz",     "pop b",   "jnz",     "jmp",     "cnz",     "push b",   "adi",     "rst 0",   "rz",      "ret",     "jz",      "illegal", "cz",      "call",    "aci",     "rst 1",
    "rnc",     "pop d",   "jnc",     "out",     "cnc",     "push d",   "sui",     "rst 2",   "rc",      "illegal", "jc",      "in",      "cc",      "illegal", "sbi",     "rst 3",
    "rpo",     "pop h",   "jpo",     "xthl",    "cpo",     "push h",   "ani",     "rst 4",   "rpe",     "pchl",    "jpe",     "xchg",    "cpe",     "illegal", "xri",     "rst 5",
    "rp",      "pop psw", "jp",      "di",      "cp",      "push psw", "ori",     "rst 6",   "rm",      "sphl",    "jm",      "ei",      "cm",      "illegal", "cpi",     "rst 7"
);

#------------------------------------------------------------------------------
%VARIABLES = (
# 4.1   BIOS.
    0xBE10 => [ 1, 'KEYCODE',     '  '         ],
    0xBE14 => [ 1, 'KEYSTAT',     ' '        ],
    0xBE1B => [ 1, 'DISP_OUT',    './. '         ],
    0xBE1D => [ 1, 'KYB_MODE',    ' '            ],
    0xBE1E => [ 1, 'KYB_BEEP',    './.  .'       ],
    0xBE30 => [ 2, 'TXT_ADDR',    '  '        ],
    0xBE31 => [ 0, 'TXT_ADDR+1',  '   + 1'    ],
    0xBE32 => [ 1, 'COL#',        'X ... '      ],
    0xBE33 => [ 1, 'ROW#',        'Y ... '      ],
    0xBE36 => [ 1, 'TXT_COLOR',   ' '                 ],
    0xBE38 => [ 1, 'BORDER',      '  '       ],
    0xBE39 => [ 1, 'SCRL_LOCK',   ' ./.'        ],
    0xBE3C => [ 1, 'CURSOR',      ' '           ],
    0xBE3D => [ 1, 'CURSOR*',     '.  '      ],
    0xBE3E => [ 1, 'LOC_COL',     '. X . '          ],
    0xBE3F => [ 1, 'LOC_ROW',     '. Y . '          ],
    0xBE40 => [ 1, 'PRN_SHIFT',   '  .'       ],
    0xBE41 => [ 1, 'PRN_XOR',     '. PRN ./.'     ],
    0xBE50 => [ 1, 'X1',          'X  '           ],
    0xBE51 => [ 1, 'Y1',          'Y  '           ],
    0xBE52 => [ 1, 'GRF_COLOR',   ' . '        ],
    0xBE57 => [ 1, 'X2',          'X  '            ],
    0xBE58 => [ 1, 'Y2',          'Y  '            ],
    0xBE80 => [ 2, 'WR0PERIOD',   '. ."0"'       ],
    0xBE81 => [ 0, 'WR0PERIOD+1', '. ."0"'       ],
    0xBE82 => [ 2, 'WR1PERIOD',   '. ."1"'       ],
    0xBE83 => [ 0, 'WR1PERIOD+1', '. ."1"'       ],
    0xBE84 => [ 1, 'PILOT_DUR',   ' '         ],
    0xBE85 => [ 1, '_ML_PAR_1_',  '   (. .)' ],
    0xBE86 => [ 1, '_ML_PAR_2_',  '   (. .)' ],
    0xBE87 => [ 1, 'FILE_TYPE',   '. '      ],
    0xBE88 => [ 2, 'BASFL_ORG',   ' BASIC-'          ],
    0xBE89 => [ 0, 'BASFL_ORG+1', ' BASIC- + 1'      ],
    0xBE8A => [ 2, 'BASFL_END',   ' BASIC-'           ],
    0xBE8B => [ 0, 'BASFL_END+1', ' BASIC- + 1'       ],
    0xBE8C => [ 6, 'SAVE_NAME',   '  /'     ],
    0xBE8D => [ 0, 'SAVE_NAME+1', '  / + 1' ],
    0xBE8E => [ 0, 'SAVE_NAME+2', '  / + 2' ],
    0xBE8F => [ 0, 'SAVE_NAME+3', '  / + 3' ],
    0xBE90 => [ 0, 'SAVE_NAME+4', '  / + 4' ],
    0xBE91 => [ 0, 'SAVE_NAME+5', '  / + 5' ],
    0xBE92 => [ 6, 'LOAD_NAME',   '  '         ],
    0xBE93 => [ 0, 'LOAD_NAME+1', '   + 1'     ],
    0xBE94 => [ 0, 'LOAD_NAME+2', '   + 2'     ],
    0xBE95 => [ 0, 'LOAD_NAME+3', '   + 3'     ],
    0xBE96 => [ 0, 'LOAD_NAME+4', '   + 4'     ],
    0xBE97 => [ 0, 'LOAD_NAME+5', '   + 5'     ],
    0xBEA3 => [ 1, 'AUTOSTART',   ' '          ],
    0xBEA4 => [ 2, 'CODFL_ORG',   ' CODE-'           ],
    0xBEA5 => [ 0, 'CODFL_ORG+1', ' CODE- + 1'       ],
    0xBEA6 => [ 2, 'CODFL_END',   ' CODE-'            ],
    0xBEA7 => [ 0, 'CODFL_END+1', ' CODE- + 1'        ],
    0xBEA9 => [ 2, 'CODFL_RUN',   ' CODE-'            ],
    0xBEAA => [ 0, 'CODFL_RUN+1', ' CODE- + 1'        ],
    0xBEAB => [ 2, 'OFFSET',      ' CODE-'         ],
    0xBEAC => [ 0, 'OFFSET+1',    ' CODE- + 1'     ],
    0xBEC0 => [ 1, 'PALETTE',     ''                     ],
    0xBEC1 => [ 1, 'GROUND',      ' '                   ],
    0xBEF1 => [ 2, 'LOAD_ERR', '..  ./'         ],
    0xBEF2 => [ 0, 'LOAD_ERR+1', '..  ./ + 1'   ],
    0xBEF3 => [ 1, 'PRN_OUT',     './. '          ],

# 4.2    BASIC.
    0x0243 => [ 2, 'BAS_PROG',    '.. BAS-'      ],
    0x0244 => [ 0, 'BAS_PROG+1',  '.. BAS- + 1'  ],
    0x0245 => [ 2, 'BAS_VARS',    '.. BAS-.'      ],
    0x0246 => [ 0, 'BAS_VARS+1',  '.. BAS-. + 1'  ],
    0x0247 => [ 2, 'BAS_ARRS',    '.. BAS-'       ],
    0x0248 => [ 0, 'BAS_ARRS+1',  '.. BAS- + 1'   ],
    0x0249 => [ 2, 'BAS_FREE',    '.. . '       ],
    0x024A => [ 0, 'BAS_FREE+1',  '.. .  + 1'   ]
);

#------------------------------------------------------------------------------
%PROCS = (
    0xF800 => 0xBFEB, 0xBFEB => 0xD7D9, 0xD7D9 => [ 'BASIC' ],
    0xF803 => 0xBFEE, 0xBFEE => 0xE800, 0xE800 => [ 'KEY',
        '2.2.1        '
    ],
    0xF806 => 0xBFF1, 0xBFF1 => 0xE843, 0xE843 => [ 'KY*',
        '2.2.2      '
    ],
    0xF809 => 0xBFF4, 0xBFF4 => 0xDF68, 0xDF68 => [ 'TTY',
        '2.2.4    '
    ],
    0xF80C => 0xBFF7, 0xBFF7 => 0xE236, 0xE236 => [ 'PRN',
        '2.2.5    '
    ],
    0xF80F => 0xBFFA, 0xBFFA => 0xE222, 0xE222 => [ 'LST',
        '2.2.6     / '
    ],
    0xF812 => 0xBFFD, 0xBFFD => 0xE888, 0xE888 => [ 'STK',
        '2.2.3  '
    ],
    0xF815 => 0xFF90, 0xFF90 => [ 'CHECKSUM',
        '2.2.10      '
    ],
    0xF818 => 0xF067, 0xF067 => [ 'VADDR',
        '2.6.3     '
    ],
    0xF81B => 0xDEBF, 0xDEBF => [ 'BEEP',
        '2.6.1   '
    ],
    0xF81E => 0xDE94, 0xDE94 => [ 'SOUND',
        '2.6.2   '
    ],
    0xF821 => 0xF048, 0xF048 => [ 'PSET',
        '2.3.2    '
    ],
    0xF824 => 0xF112, 0xF112 => [ 'LINE',
        '2.3.4    '
    ],
    0xF827 => 0xF1A5, 0xF1A5 => [ 'BOX',
        '2.3.5    '
    ],
    0xF82A => 0xF1DA, 0xF1DA => [ 'FIL_BOX',
        '2.3.6   '
    ],
    0xF82D => 0xDF02, 0xDF02 => [ 'LOCATE',
        '2.2.11  '
    ],
    0xF830 => 0xE774, 0xE774 => [ 'PAINT',
        '2.3.7   '
    ],
    0xF833 => 0xDE3C, 0xDE3C => [ 'COLOR',
        '2.3.8   '
    ],
    0xF836 => 0xEBD0, 0xEBD0 => [ 'CLS',
        '2.3.1  '
    ],

    0xE4A4 => [ 'TEXT',
        '2.2.7     '
    ],
    0xFFD6 => [ '_HEX_A_',
        '2.2.8      A   '
    ],
    0xFFD1 => [ '_HEX_HL_',
        '2.2.9       HL   '
    ],

    0xF020 => [ 'PRESET',
        '2.3.3  '
    ],
    0xE627 => [ 'COPY',
        '2.3.9    '
    ],

    0xE2BE => [ 'WR_0',
        '2.4.1   "0"'
    ],
    0xE2C5 => [ 'WR_1',
        '2.4.2   "1"'
    ],
    0xE2D5 => [ '_WR_MEANDR_',
        '2.4.3   .'
    ],
    0xE42B => [ 'WR_PILOT',
        '2.4.4  -'
    ],
    0xE437 => [ 'WR_BYTE',
        '2.4.5  '
    ],
    0xDD86 => [ 'WR_WORD',
        '2.4.6  '
    ],
    0xE3E4 => [ 'WR_HEAD',
        '2.4.7   '
    ],
    0xDD31 => [ 'BSAVE',
        '2.4.8   '
    ],

    0xE4D0 => [ 'RD_PILOT',
        '2.5.1  -'
    ],
    0xE4BE => [ 'RD_BYTE',
        '2.5.2  '
    ],
    0xDDCA => [ 'RD_WORD',
        '2.5.3  '
    ],
    0xE443 => [ 'RD_HEAD',
        '2.5.4   '
    ],
    0xDDBC => [ 'RD_BLOCK',
        '2.5.5  '
    ],
    0xDD94 => [ 'BLOAD',
        '2.5.6   '
    ],

    0xEBBC => [ '_FILL_VRAM_',
        '2.6.4    .'
    ],
    0xEBA5 => [ '_FILL_SCREEN_',
        '2.6.5    .'
    ],
    0xEBE4 => [ '_CUR_HOME_',
        '2.6.6    .'
    ],
    0xE11F => [ 'MOVE',
        '2.6.7  '
    ],
    0xE12A => [ 'FILL',
        '2.6.8    '
    ],
    0xFFBD => [ '_HNIB2DEC_',
        '2.6.9    HEX  DEC'
    ],
    0xFFDF => [ '_NIB2SCRHEX_',
        '2.6.10      HEX'
    ],
    0xFF24 => [ '_SUM_X_',
        '2.6.11  DE  A (  A  )'
    ],
    0xE425 => [ '_DE=HL_',
        '2.6.12   DE=HL'
    ],
    0xDB63 => [ '_FILL_SPACES_',
        '2.6.13  '
    ],

    0xE6AA => [ 'P_E6AA',
        '   ESC+chr$(C)'
    ],
    0xE6A3 => [ 'P_E6A3',
        '   chr$(C)'
    ]
);

sub get_proc {
    my $p = $_[0];
    return $PROCS{$PROCS{$PROCS{$p}}} if defined $PROCS{$PROCS{$PROCS{$p}}};
    return $PROCS{$PROCS{$p}} if defined $PROCS{$PROCS{$p}};
    return $PROCS{$p} if defined $PROCS{$p};
    return undef();
}

#------------------------------------------------------------------------------
# This is AEROCOBR.LVT from warez-d4.rar\ARCHIVE\GAMES.BIN\AEROCOBR.LVT

$fname = ($ARGV[0] or "AEROCOBR.LVT");
$tname = ($ARGV[1] or "AEROCOBR.DEF");
$code_page = ($ARGV[2] or 1251);

#------------------------------------------------------------------------------
open F,"<$fname" or die "$fname - $!";
binmode F; read F, $d, -s F; close F;

($sign, $type, $name, $beg, $end, $run) = unpack "A9CA6vvv", $d;
die "Not a BLOAD file" unless $type == 0xD0;

open F, "<$tname" or die "$tname - $!";
while(<F>) {
    chomp;
    next if /^\s*$/ or /^[#;]/;

    $META = $1, next if /^\@(\S+)/;

    @a = split /\s+/, $_;
    if ($a[1] eq '') { $META{hex $a[0]} = $META } else {
        $LABELS{hex $a[0]} = $a[1];
        $STRINGS{hex $a[0]} = $a[1] if $a[2] ne '';
    }
}
close F;

#------------------------------------------------------------------------------
$offs = 0x16;

$body = substr $d, $offs;
$cksum = 0; $cksum += ord $1 while $body =~ m{(.)}sg;

printf <<HEADER, $name, $beg, $end, $run, $cksum;
<html>
<body>
<pre>
;
; BLOAD "%s" %04X-%04X,%04X   CkSum: %06X
;
HEADER

while ($offs < length($d)) {
    $addr = $offs - 0x16 + $beg;

    if ($STRINGS{$addr}) {
        for ($sz = 0; $sz < 256; $sz++) {
            if (defined $LABELS{$addr+$sz+1}
            or substr($d, $offs + $sz, 1) eq "\x00") {
                $sz++; last;
            }
        }
        $insn = "DB " . ccode(xcode(substr($d, $offs, $sz)));
    }
    else {
        $code = ord(substr($d, $offs, 1));
        $sz = $INSTR_SZ[$code];
        $insn = $INSTR[$code];
        if ($sz == 2) {
            if ($META{$addr} eq 'CHAR') {
                $insn .= sprintf " '%s'", xcode(substr($d, $offs+1, 1));
            }
            elsif ($META{$addr} eq 'UNSIGNED') {
                $insn .= sprintf " %u", unpack "C", substr($d, $offs+1, 1);
            }
            else {
                $insn .= sprintf " 0%02Xh", unpack "C", substr($d, $offs+1, 1);
            }
        }
        elsif ($sz == 3) {
            $param = unpack "v", substr($d, $offs+1, 2);
            if ($LABELS{$param}) {
                $insn .= " " . $LABELS{$param};
            }
            elsif ($VARIABLES{$param}) {
                $insn .= sprintf " \$%s ; %s (0%04Xh)",
                    $VARIABLES{$param}[1],
                    $VARIABLES{$param}[2],
                    $param;
            }
            elsif ($PROCS{$param}) {
                $proc = get_proc($param);
                $insn .= sprintf " \@%s ; %s (0%04Xh)",
                    $proc->[0],
                    $proc->[1],
                    $param;
            }
            else {
                $insn .= sprintf " 0%04Xh", $param;
            }
        }
        else {
        }

# align 1st parameter
        unless ($insn =~ s{^(\S+) }{substr($1 . (' ' x 10), 0, 10)}e) {
            $insn = substr($insn . (' ' x 10), 0, 10);
        }

# align comment
        $insn =~ s{^(.*);}{substr($1 . (' ' x 30), 0, 30).";"}e;

# align everything
        $insn = substr($insn . (" " x 29), 0, 29)
            if length($insn) < 29;
    }

    if ($sz <= 3) {
        $bytes = unpack("H*", substr($d, $offs, $sz)) . ("-" x 4);
        $bytes =~ s{(..)}{\1 }g;
        $chars = dcode(xcode(substr($d, $offs, $sz))) . ("-" x 2);
    }
    else {
        $bytes = "-" x 8;
        $chars = "-" x 3;
    }
    $bytes = substr($bytes, 0, 8);
    $chars = substr($chars, 0, 3);

    $label = "";
    $label = $LABELS{$addr} . ":" if defined $LABELS{$addr};
    $label = substr($label . (" " x 10), 0, 10);

    $line = sprintf "%04X:  %s  %s  %s %s",
        $addr, $bytes, $chars, $label, $insn;

    $line =~ s{\s+$}{};
#   $line =~ s{\&}{\&amp;}sg;
#   $line =~ s{\>}{\&gt;}sg;
#   $line =~ s{\<}{\&lt;}sg;
#   $line =~ s{(?<=.)0([0-9A-F]{4})h}{"<a href='#$1'>0$1h</a>"}esg;
#   $line =~ s{^(\S+):}{"<a name='$1'>$1</a>:"}esg;
    print $line, "\n";

    $count++;
    $offs += $sz;
}

printf <<FOOTER, $count;

; lines: %d
</pre>
</body>
</html>
FOOTER
__END__
